"""
This module contains the activation functions 
"""
import numpy as np
import mygrad as mg
import mygrad.nnet as nnet

def ReLU(x: np.array) -> np.array:
    return mg.maximum(0, x)

def softmax(x: mg.Tensor, axis=-1) -> mg.Tensor:
    x_shifted = x - mg.max(x, axis=axis, keepdims=True)
    e_x = mg.exp(x_shifted)
    return e_x / mg.sum(e_x, axis=axis, keepdims=True)

def sigmoid(x: mg.Tensor) -> mg.Tensor:
    return nnet.activations.sigmoid(x)

def tanh(x: mg.Tensor) -> mg.Tensor:
    return nnet.activations.tanh(x)



"""
This module contains layers.
"""
from typing import List, Dict, Tuple, Optional
from abc import ABC
from numbers import Integral  # Max Pooling
import inspect

from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from mygrad.nnet.layers.utils import sliding_window_view

import numpy as np
import mygrad as mg
from mygrad import nnet

from . import activations

names: Dict[str, int] = {}


class Layer(ABC):
    """
    :param List[mg.Tensor] trainable_params: A list that consists of the parameters that
        will be trained by the optimiser. Pass an empty list to indicate that it has no
        trainable parameters

    This the base class of all layers that gives them a unique name
    if they have trainable parameters and if they don't they will just
    use their type name without counting.
    """

    def __init__(self, trainable_params: List[mg.Tensor],) -> None:
        self.type = self.__class__.__name__
        # self.total_params = total_params
        self._config = self._get_current_config()
        self.trainable_params: List[mg.Tensor] = trainable_params

        if trainable_params != []:
            # In case of the layer doesn't have any trainable parameters.
            # We won't add numbers for it

            if (number := names.get(self.__class__.__name__)):
                self.name = self.__class__.__name__ + " " + str(number)
                names[self.__class__.__name__] += 1
            else:

                self.name = self.__class__.__name__ + " 1"
                names[self.__class__.__name__] = 2  # Because we already used `1`
            for i, parameter in enumerate(trainable_params):
                if isinstance(parameter, np.ndarray):
                    parameter = mg.tensor(parameter, constant=False)
                    self.trainable_params[i] = parameter
                elif isinstance(parameter, mg.Tensor):
                    continue
                else:
                    raise TypeError(f"The parameter of {type(parameter)} at index {i}"
                                    f" must be np.ndarray or mygrad.Tensor")
        else:

            self.name = self.__class__.__name__

    def __call__(self, *args, **kwargs) -> mg.Tensor:
        """
        This operator should perform a forward propagation
        """  # This behaviour will change
        self.forward(*args, **kwargs)

    # pylint: disable-next=C0103
    def forward(self, X: mg.Tensor) -> mg.Tensor:
        """
        :param mg.Tensor | np.ndarray X: The data that should be forward propagated
        :return: The prediction of the layer
        :rtype: mg.Tensor

        this method performs forward propagation
        """
    @mg.no_autodiff
    def predict(self, *args, **kwargs) -> mg.Tensor:
        """
        :param mg.Tensor X: The data of forward pass
        :return: The prediction of the layer
        :rtype: mg.Tensor

        This function should in inference not in training because
        it doesn't track the gradients
        """
        return self.forward(*args, **kwargs)

    def null_grad(self):
        """
        This function resets the gradients of parameters.
        This method should not be modified by children

        This function should be used we want to be sure the gradients don't stack up
        :return: This method doesn't return anything
        """
        for parameter in self.trainable_params:
            parameter.null_grad()

    def output_shape(self, prev_shape: Optional[Tuple[int]] = ()) -> Tuple[int]:
        """
        A simple method that gets the shape of layer's output

        This function should be overloaded in order to get the shape
        but still can work on it's own but it'll need previous shapes
        in order to work properly
        """
        return self.predict(np.zeros((1, *prev_shape))).shape[1:]

    def total_params(self):
        """This method can calculate the number of parameters in each layer

        :rtype: int
        """
        params = 0
        for parameter in self.trainable_params:
            params += np.array(parameter.shape).prod()
        return params

    def _get_current_config(self) -> dict:
        """
        An internal helper to automatically capture the constructor's arguments.
        This is a bit of advanced Python ("introspection").
        """
        # Get the class of the object being created (e.g., Dense, Conv2D)
        cls = self.__class__
        # Get the signature of its __init__ method
        init_signature = inspect.signature(cls.__init__)

        # Get the names of the arguments
        arg_names = [p.name for p in init_signature.parameters.values() if p.name != 'self']

        # Create a dictionary of these arguments and their current values
        config = {"class_name": self.type}
        for name in arg_names:
            if hasattr(self, name):
                value = getattr(self, name)
                # Make it JSON-safe
                if callable(value):
                    config[name] = value.__name__
                elif isinstance(value, (int, str, float, bool, list, tuple, type(None))):
                    config[name] = value
        return config

    def get_config(self) -> dict:
        """Public method to get the stored configuration."""
        return self._config

    def get_weights(self) -> List[np.ndarray]:
        """
        returns a list of weights of them model.
        """
        return [p.data for p in self.trainable_params]

    def set_weights(self, weights: List[np.ndarray]):
        """
        Sets the weights of them model.

        :param List[np.ndarray] weights: This should be the return value of get_weights
        """
        for param, weight_array in zip(self.trainable_params, weights):
            param.data = weight_array

    def populate_self(self, population: dict) -> None:
        """
        This function used in inhereitance in order to
        to get the arguments in and save ``self`` to enable automatic saving

        :param dict population: the argument should be ``local()``
        """
        for key, value in population.items():
            setattr(self, key, value)


class Dense(Layer):
    """
    :param int inputs: The size of inputs.
    :param int params: The size of parameters. It is the size of the output.
    :param activation: The activation function that will be used with the layer.
        Any function or object with ``__call__()`` method.
    :param bool use_bias: Whether to have a bias or not.
    :param dtype: The data type of the parameters of the layers also using
        data types from **MyGrad** is preferred over NumPy

    A simple dense layer.
    """

    # pylint: disable-next=R0913
    def __init__(self, inputs: int, params: int, activation,
                 use_bias: bool = True, dtype=mg.float32) -> None:
        self.inputs = inputs
        self.params = params
        he_stddev = np.sqrt(2. / np.array(inputs).sum())
        self.use_bias = use_bias
        self.activation = activation
        self.out_sh = tuple([params]) if isinstance(params, int) else params
        inputs = tuple([inputs]) if isinstance(inputs, int) else inputs
        self.weights = mg.tensor(np.random.randn(*inputs, *self.out_sh) * he_stddev, constant=False,
                                 dtype=dtype)
        if self.use_bias:
            self.bias = mg.tensor(np.random.randn(1, *self.out_sh), constant=False,
                                  dtype=dtype)

            super().__init__([self.weights, self.bias])
        else:

            super().__init__([self.weights])

    def forward(self, X: np.array):
        """
        :param mg.Tensor X: the inputs of the layer
        :return mg.Tensor: The layer predictions
        """
        output = mg.matmul(X, self.weights)
        if self.use_bias:
            output += self.bias
        return self.activation(output)

    def output_shape(self, prev_shape: Optional[Tuple[int]] = None) -> Tuple[int]:
        return self.out_sh

# Well class _MaxPoolND taken from MyGrad.nnet implementation but it doesn't use
# Floor division so I had to create mine with floor division


class _MaxPoolND(Operation):
    def __call__(self, x, pool, stride):
        self.variables = (x,)  # data: ((N0, ...), C0, ...)
        x = x.data

        assert isinstance(pool, (tuple, list, np.ndarray)) and all(
            i >= 0 and isinstance(i, Integral) for i in pool
        )
        pool = np.asarray(pool, dtype=int)
        assert all(i > 0 for i in pool)
        assert x.ndim >= len(
            pool
        ), "The number of pooled dimensions cannot exceed the dimensionality of the data."

        stride = (
            np.array([stride] * len(pool))
            if isinstance(stride, Integral)
            else np.asarray(stride, dtype=int)
        )
        assert len(stride) == len(pool) and all(
            s >= 1 and isinstance(s, Integral) for s in stride
        )
        # pylint: disable-next= W0201
        self.pool = pool  # (P0, ...)

        self.stride = stride  # (S0, ...) # pylint: disable= W0201

        num_pool = len(pool)
        num_no_pool = x.ndim - num_pool

        x_shape = np.array(x.shape[num_no_pool:])
        w_shape = pool
        # MODIFIED: BY ME `//` instead of `/`
        out_shape = (x_shape - w_shape) // stride + 1

        if not all(i.is_integer() and i > 0 for i in out_shape):
            msg = "Stride and kernel dimensions are incompatible: \n"
            msg += f"Input dimensions: {(tuple(x_shape))}\n"
            msg += f"Stride dimensions: {(tuple(stride))}\n"
            msg += f"Pooling dimensions: {(tuple(w_shape))}\n"
            raise ValueError(msg)

        pool_axes = tuple(-(i + 1) for i in range(num_pool))

        # (G0, ...) is the tuple of grid-positions for placing each window (not including stride)
        # sliding_window_view(x): ((N0, ...), C0, ...)          -> (G0, ..., (N0, ...), P0, ...)
        # max-pool:               (G0, ..., (N0, ...), P0, ...) -> (G0, ..., (N0, ...))
        maxed = sliding_window_view(
            x, self.pool, self.stride).max(axis=pool_axes)
        axes = tuple(range(maxed.ndim))

        # (G0, ..., (N0, ...)) -> ((N0, ...), G0, ...)
        out = maxed.transpose(axes[-num_no_pool:] + axes[:-num_no_pool])
        return out if out.flags["C_CONTIGUOUS"] else np.ascontiguousarray(out)

    def backward_var(self, grad, index, **kwargs): # pylint: disable= R0914
        """Parameters
        ----------
        grad : numpy.ndarray, shape=((N0, ...), G0, ...),
        index : int"""
        var = self.variables[index]
        x = var.data
        num_pool = len(self.pool)

        sl = sliding_window_view(x, self.pool, self.stride)
        grid_shape = sl.shape
        maxed = sl.reshape(*sl.shape[:-num_pool], -1).argmax(-1)
        axes = tuple(range(maxed.ndim))

        # argmax within a given flat-window
        maxed = maxed.transpose(
            axes[num_pool:] + axes[:num_pool]
        )  # ((N0, ...), G0, ...)

        # flat-index offset associated with reshaped window within `x`
        row_major_offset = tuple(np.cumprod(
            x.shape[-num_pool:][:0:-1])[::-1]) + (1,)

        # flat index of argmax, updated based on position within window, according to shape of `x`
        in_window_offset = sum(
            ind * off
            for ind, off in zip(np.unravel_index(maxed, self.pool), row_major_offset)
        )

        # flat-index of strided window placement, relative to `x`
        window_offset = sum(
            ind * s * off
            for ind, s, off in zip(
                np.indices(grid_shape[:num_pool]
                           ), self.stride, row_major_offset
            )
        )

        # indices required to traverse pool-axis-flattened array
        # ((N0, ...) G0*...)
        flat_grid_shape = (
            *maxed.shape[:-num_pool], np.prod(maxed.shape[-num_pool:]))
        index = np.indices(flat_grid_shape)

        # update trailing indices to traverse location of max entries within pooled axes
        index[-1] = (in_window_offset + window_offset).reshape(
            *flat_grid_shape[:-1], -1
        )

        # accumulate gradient within pool-axis-flattened dx, then reshape to match shape of `x`
        dx = np.zeros(x.shape[:-num_pool] + (np.prod(x.shape[-num_pool:]),))
        np.add.at(dx, tuple(index), grad.reshape(*x.shape[:-num_pool], -1))
        return dx.reshape(x.shape)


def _max_pool(
    x,
    pool,
    stride,
    *,
    constant=None,
) -> Tensor:
    # pylint: disable-next= W0212
    return Tensor._op(_MaxPoolND, x, op_args=(pool, stride), constant=constant)


class Conv2D(Layer):
    """
    :param int input_channel: the expected number of input channels.
    :param int output_channel: The number of channels should the layer output.
    :param int kernel_size: the size of the kernel.
    :param activation: The activation function that will be used with the layer.
        Any function or object with ``__call__()`` method.
    :param bool use_bias: Whether to have a bias or not.

    :class:`helixnet.layers.Conv2D` assumes input data is of
        shape **(N, C_in, H, W)**:

    **N**: batch size

    **C_in**: number of input channels

    **H**: height of the input feature map

    **W**: width of the input feature map
    """

    def __init__(self, input_channels: int, output_channels: int, kernel_size,
                 stride=1, padding=0, activation=None, use_bias: bool = True):

        self.populate_self(locals())
        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)

        # Xavier/Glorot initialization for weights (filters)
        # Shape: (C_out, C_in, K_H, K_W)
        weight_shape = (output_channels, input_channels, *kernel_size)
        self.weights = mg.tensor(
            np.random.randn(*weight_shape)
            * np.sqrt(2.
                    / (input_channels * kernel_size[0] * kernel_size[1]))
        )

        self.use_bias = use_bias
        self.activation = activation
        self.kernel_size = kernel_size
        if self.use_bias:
            # Bias has one value per output channel
            self.bias = mg.tensor(np.zeros(output_channels))
            super().__init__([self.weights, self.bias])
        else:
            self.bias = None
            super().__init__([self.weights])

        self.stride = stride
        self.padding = padding

    def forward(self, X: mg.Tensor) -> mg.Tensor:
        """
        Performs a forward pass.

        Also the images needs t be in the following shape
        (batch_size, color_channels, length, width)
        """
        conv_result = nnet.conv_nd(X, self.weights, stride=self.stride,
                                   padding=self.padding)

        if self.use_bias:
            # Reshape bias for broadcasting: (C_out,) -> (1, C_out, 1, 1)
            conv_result = conv_result + self.bias.reshape(1, -1, 1, 1)
        return self.activation(conv_result)


class Flatten(Layer):
    """
    A simple flatten layer that turns it's inputs into a flat layer
    """

    def __init__(self):
        super().__init__([])

    def forward(self, X: mg.Tensor) -> mg.Tensor:
        """
        :param mg.Tensor X: The tensor that will be flattened
        :return mg.Tensor: A flat tensor

        Takes an input of shape (N, C, H, W) and flattens it
        to a shape of (N, C*H*W).
        """

        # The first dimension (batch size, N) is preserved.
        # The rest of the dimensions are flattened.

        return X.reshape(X.data.shape[0], -1)

    def output_shape(self, prev_shape: Tuple[int] = ()) -> Tuple[int]:
        return (np.array(prev_shape)[0:].prod(),)


class MaxPooling2D(Layer):
    """
    :param int | Tuple[int, int] pool_size: the pool size can be integer for square
        pools and can be a tuple for rectangular tuples

    A layer to perform max pooling over a 4D input (No_samples, Channels, Height, Width).
    """

    def __init__(self, pool_size, stride=None):
        self.populate_self(locals())
        super().__init__([])
        if isinstance(pool_size, int):
            self.pool_size = (pool_size, pool_size)
        else:
            self.pool_size = pool_size

        # If stride is not specified, it defaults to the pool_size
        if stride is None:
            self.stride = self.pool_size
        elif isinstance(stride, int):
            self.stride = (stride, stride)
        else:
            self.stride = stride

    def forward(self, X: mg.Tensor) -> mg.Tensor:
        """
        Applies the max pooling operation.
        """
        return _max_pool(X, self.pool_size, self.stride)


class LSTMCell(Layer):
    """
    A single cell of an LSTM. Performs the computation for one timestep.
    """

    def __init__(self, input_size: int, hidden_size: int):
        self.input_size = input_size
        self.hidden_size = hidden_size

        # The input to the gates is the concatenation of the previous hidden state
        # and the current input, so its size is hidden_size + input_size.
        concat_size = hidden_size + input_size

        # We create one large weight matrix for all 4 gates (input, forget, candidate, output)
        self.weights_all = mg.tensor(
            np.random.randn(concat_size, 4 * hidden_size) *
            np.sqrt(2. / concat_size), dtype=mg.float32)

        # We also create one large bias vector for all 4 gates.
        self.bias_all = mg.tensor(
            np.zeros((1, 4 * hidden_size)), dtype=mg.float32)
        # A common practice is to initialize the forget gate bias to 1.0 to encourage remembering.
        self.bias_all.data[:, hidden_size: 2 * hidden_size] = 1.0

        super().__init__([self.weights_all, self.bias_all])

    def forward(self, x_t: mg.Tensor, h_prev: mg.Tensor, C_prev: mg.Tensor):
        """
        Performs a forward pass for a single timestep.


        :param x_t (mg.Tensor): Input for the current timestep, shape (N, input_size).
        :param h_prev (mg.Tensor): Hidden state from the previous timestep, shape (N, hidden_size).
        :param C_prev (mg.Tensor): Cell state from the previous timestep, shape (N, hidden_size).

        :return Tuple[mg.Tensor, mg.Tensor]: The new hidden state (h_next) and cell state (C_next).
        """
        # Concatenate previous hidden state and current input
        concat_input = mg.concatenate([h_prev, x_t], axis=1)

        # Perform a single matrix multiplication for all gates
        gate_calcs = mg.matmul(concat_input, self.weights_all) + self.bias_all

        # This is the correct and idiomatic way in MyGrad.
        hs = self.hidden_size
        f_calc = gate_calcs[:, :hs]
        i_calc = gate_calcs[:, hs: 2 * hs]
        g_calc = gate_calcs[:, 2 * hs: 3 * hs]
        o_calc = gate_calcs[:, 3 * hs:]  # Slicing to the end is robust

        # Apply activation functions to each gate
        f_t = activations.sigmoid(f_calc)
        i_t = activations.sigmoid(i_calc)
        g_t = activations.tanh(g_calc)
        o_t = activations.sigmoid(o_calc)

        # Calculate the new cell state and hidden state
        C_next = (f_t * C_prev) + (i_t * g_t)
        h_next = o_t * activations.tanh(C_next)

        return h_next, C_next


class LSTMLayer(Layer):
    """
    An LSTM layer that processes a sequence of inputs by unrolling an LSTMCell
    over time.
    """

    def __init__(self, input_size: int, hidden_size: int, return_sequences: bool = True):
        self.hidden_size = hidden_size
        self.return_sequences = return_sequences

        # The LSTM layer contains a cell that holds the parameters
        self.cell = LSTMCell(input_size, hidden_size)

        # Expose the cell's parameters as the layer's parameters for the optimizer
        self.weights = self.cell.weights_all
        self.bias = self.cell.bias_all
        # LSTM layer doesn't need its own use_bias attribute, it's handled by the cell
        self.use_bias = True
        super().__init__([self.weights, self.bias])

    def forward(self, X: mg.Tensor) -> mg.Tensor:
        """
        Processes a sequence of inputs.

        :param mg.Tensor X: The input sequence, shape (N, seq_len, input_size).
        :return mg.Tensor: The sequence of hidden states, shape (N, seq_len, hidden_size),
                       or the final hidden state, shape (N, hidden_size).
        """
        batch_size, seq_len, _ = X.shape

        # Initialize hidden state and cell state with zeros
        h_prev = mg.tensor(
            np.zeros((batch_size, self.hidden_size)), dtype=mg.float32)
        C_prev = mg.tensor(
            np.zeros((batch_size, self.hidden_size)), dtype=mg.float32)

        # List to store the hidden states from each timestep
        outputs = []

        # Unroll the cell over the time dimension (seq_len)
        for t in range(seq_len):
            # Get the input for the current timestep
            x_t = X[:, t, :]

            # Run the cell
            h_next, C_next = self.cell.forward(x_t, h_prev, C_prev)

            # Store the output hidden state
            outputs.append(h_next)

            # Update states for the next iteration
            h_prev, C_prev = h_next, C_next

        if self.return_sequences:
            # Stack all hidden states to create a single output tensor
            # We need to reshape each h_next to (N, 1, hidden_size) before concatenating
            reshaped_outputs = [out.reshape(
                batch_size, 1, self.hidden_size) for out in outputs]
            return mg.concatenate(reshaped_outputs, axis=1)
        else:
            return outputs[-1]


class Embedding(Layer):
    """
    Word embedding layer

    :param int vocab_size: The size of vocabulary
    :param int dim: the number of output dimensions
    """

    def __init__(self, vocab_size, dim) -> None:
        self.vocab_size = vocab_size
        self.dim = dim

        self.weight = mg.tensor((np.random.rand(vocab_size, dim) - 0.5) / dim,
                                constant=False)

        super().__init__([self.weight])

    def forward(self, X: mg.Tensor):
        return self.weight[X]


class InputShape(Layer):
    """
    A very simple layer designed just for model designing
    where you might need in order to determine the model shapes and
    it can make sure the input shape is correct where it the shape of data is
    (X, D1, D2, D3, ... ) where it ignores the the first dimension because it is
    the number of samples.
    """

    def __init__(self, shape: Tuple[int],
                 ensure_shape: Optional[bool] = True) -> None:
        self.shape = tuple(shape)
        self.ensure_shape = ensure_shape
        super().__init__([])

    def forward(self, X: mg.Tensor) -> mg.Tensor:
        """Just returns the inputs"""
        if self.ensure_shape and self.shape != X.shape[1:]:
            raise ValueError(f"The input shape of {X.shape} doesn't match "
                             f"the desired {self.shape}")
        return X

    def output_shape(self, prev_shape: Optional[Tuple[int]] = None) -> Tuple[int]:
        """returns the input shape of layers"""
        return self.shape


class Dropout(Layer):
    """
    A dropout layer

    :param float proba: The percentage of inactive neurons
    """

    def __init__(self, proba: float) -> None:
        self.proba = proba
        super().__init__([])

    def forward(self, X):
        keep_proba = 1 - self.proba
        mask = np.random.binomial(1, keep_proba, size=X.data.shape)
        # The division scales up the active neurons to compensate for the dropped ones
        return (X * mask) / keep_proba

    def predict(self, X):
        """
        In the method :class:`helixnet.layers.Dropout.predict`
        the :class:`helixnet.layers.Dropout` won't perform the dropout and
        pass the inputs directly.
        """
        return X


class BatchNorm(Layer):
    """
    Performs batch normalization.

    :param Tuple[int] input_shape: The shape of the input of the data
    :param float momentum: The momentum of the layer
    :param float epsilon: A simple number for numerical stability
    """

    def __init__(self, input_shape: Tuple[int], momentum=0.99, epsilon=1e-7):
        self.weight = mg.tensor(np.random.randn(*input_shape))
        self.bias = mg.tensor(np.random.randn(*input_shape[1:]))
        super().__init__([self.weight, self.bias])

        self.momentum = momentum
        self.epsilon = epsilon

        self.running_mean = np.zeros(input_shape)
        self.running_var = np.ones(input_shape)
        self.populate_self(locals())

    def forward(self, X) -> mg.Tensor:
        # During training, use batch statistics
        batch_mean = mg.mean(X, axis=0)
        batch_var = mg.var(X, axis=0)

        # Update running averages
        self.running_mean = self.momentum * self.running_mean + \
            (1 - self.momentum) * batch_mean.data
        self.running_var = self.momentum * self.running_var + \
            (1 - self.momentum) * batch_var.data

        # Normalize with batch statistics
        normalized_x = (X - batch_mean) / mg.sqrt(batch_var + self.epsilon)

        return self.weight * normalized_x + self.bias

    def predict(self, X, *args, **kwargs) -> mg.Tensor:
        """
        Normalize the data without updating the running mean and variance

        :param mg.Tensor X: The input of the data
        """
        normalized_x = (X - self.running_mean) / \
            mg.sqrt(self.running_var + self.epsilon)
        return self.weight * normalized_x + self.bias


class DenseTranspose(Layer):
    """
    This layer can be used to tie the weights only of a dense layer useful in autoencoders

    This layer doesn't tie the bias and can create it's own or not

    :param Dense layer: The layer that its weights will be tied.
    :param activation: The activation function that will be used by the layer. But if it's ``none`` it
        will use the tied layer's activation function.
    :param bool use_bias: If you want the layer to created it's own bias.
    """

    def __init__(self, layer: Dense, activation=None, use_bias=None):
        self.weight = layer.weights.T
        if use_bias or (use_bias is None and layer.use_bias):
            self.bias = mg.tensor(np.zeros(self.weight.shape[1:]),
                                  constant=False)
            super().__init__([self.bias])
        else:
            self.bias = None
            super().__init__([])
        self.activation = layer.activation if not activation \
            else activation

    def forward(self, X) -> mg.Tensor:
        return self.activation(X @ self.weight + self.bias) if self.bias \
            is not None else self.activation(X @ self.weight)

    def output_shape(self, prev_shape: Optional[Tuple[int]] = ()) -> Tuple[int]:
        return self.weight.shape[1:]
    # TODO: Check a better implementation


def upsample_zero_insert(x, scale):
    h, w = x.shape
    up = np.zeros((h * scale, w * scale), dtype=x.dtype)
    up[::scale, ::scale] = x
    return up


class ConvTranspose2D(Layer):
    """
    Performs a 2D transpose convolution (deconvolution), used for upsampling.
    This implementation works by first upsampling the input with zero-insertion,
    and then performing a standard convolution.
    """

    def __init__(self, input_channels: int, output_channels: int, kernel_size,
                 stride=1, activation=None, use_bias: bool = True):

        self.populate_self(locals())
        if isinstance(kernel_size, int):
            self.kernel_size = (kernel_size, kernel_size)
        else:
            self.kernel_size = kernel_size

        # The weight shape for the underlying conv_nd MUST be (C_out, C_in, K, K)
        # We are essentially "tricking" a regular convolution into being a transpose one.
        weight_shape = (input_channels, output_channels, *self.kernel_size)
        self.weights = mg.tensor(
            np.random.randn(*weight_shape) * np.sqrt(2. / (input_channels * self.kernel_size[0] * self.kernel_size[1]))
        )

        self.stride = stride if isinstance(stride, int) else stride[0]
        self.activation = activation if activation is not None else (lambda x: x)
        self.use_bias = use_bias

        if self.use_bias:
            self.bias = mg.tensor(np.zeros(output_channels))
            super().__init__([self.weights, self.bias])
        else:
            self.bias = None
            super().__init__([self.weights])

    def forward(self, X: mg.Tensor) -> mg.Tensor:
        """ Performs the transpose convolution using a two-step process. """
        # Step 1: Upsample the input by inserting zeros
        if self.stride > 1:
            N, C, H, W = X.shape
            # For a stride of 2, we want one zero between each element.
            H_up, W_up = (H - 1) * self.stride + 1, (W - 1) * self.stride + 1
            upsampled_data = np.zeros((N, C, H_up, W_up), dtype=X.dtype)
            upsampled_data[:, :, ::self.stride, ::self.stride] = X.data
            X_upsampled = mg.tensor(upsampled_data)
        else:
            X_upsampled = X

        # Step 2: Perform a "full" convolution on the upsampled data
        padding = self.kernel_size[0] - 1

        # We need to swap the input/output channels for the convolution's weights
        # to correctly perform the transpose operation.
        # So we transpose the first two axes of the weights.
        transposed_weights = self.weights.transpose(1, 0, 2, 3)

        conv_result = nnet.conv_nd(X_upsampled, transposed_weights, stride=1, padding=padding)

        if self.use_bias:
            conv_result += self.bias.reshape(1, -1, 1, 1)

        return self.activation(conv_result)


class Reshape(Layer):
    """Reshapes the input tensor to the specified shape.

    :param tuple[int] target_shape: The shape you want the data to be converted to."""

    def __init__(self, target_shape):
        # target_shape does not include the batch dimension (N)
        self.target_shape = target_shape
        super().__init__([])

    def forward(self, X: mg.Tensor) -> mg.Tensor:
        # The -1 in reshape is a placeholder for the batch size (N)
        return X.reshape(-1, *self.target_shape)

    def output_shape(self, prev_shape: Optional[Tuple[int]] = ()) -> Tuple[int]:
        return self.target_shape



"""The imports of the HelixNet"""
from . import activations
from . import layers
from . import loss
from . import models
from . import optimizers
from . import io
from . import metrics

__version__ = "0.6.1"



from abc import ABC
from typing import List

import numpy as np
import mygrad as mg

from . import models
from . import layers


class Regularizer(ABC):
    """A Simple regularizer class for parameter regularization"""

    def __init__(self):
        pass

    def regularize(self, parameter: mg.Tensor) -> mg.Tensor:
        """
        :param mg.Tensor parameter: the parameter that will be regularized

        This method should be overloaded by the regularizers because this method
            will perform the calculations
        """


class Optimizer(ABC):
    """
    An Abstract class that is used by other optimisers and it also performs
    the primary training loop

    :param float lr: The learn rate of the optimizer
    :param List[Regularizer] regularizers: The regularizers that
        will be applied to the parameters by the optimizer
    :param float grad_clip: The values which the gradients will be clipped to it. Or
        pass ``None`` in order to avoid performing gradient clipping
    """

    def __init__(self, lr: float, regularizers: List[Regularizer] = None,
                 grad_clip=None) -> None:
        self.step = 1
        self.regularizers = regularizers
        self.lr = lr
        self.grad_clip_value = grad_clip

    def epoch_done(self):
        """
        A simple method that should be called after each epoch.

        This method should be called after every epoch is done in order to inform the optimiser to
        update it's parameters like weight decay
        """
        self.step += 1

    def optimize_param(self, parameter: mg.Tensor) -> None:
        """
        This function takes parameters one by one and must be
        inherited by the children and overload it with the update logic

        :param parameter (mg.Tensor): The parameter itself
        """

    def get_current_lr(self) -> float:
        """
        :return: The learn rate
        :rtype: float

        This method returns the learn rate with respect to the current step
        """

    def optimize(self, model: models.Sequential, loss: mg.Tensor) -> None:
        """
        This method trains models and calls optimise_param
            for every parameter in the layer and it's called
            when the training happens
        Also if there any parameter in the model that doesn't
            have any gradients it will skip it.

        :param (models.Sequential) model: The model that will be trained
        :param mg.Tensor loss: The loss of the model
        """
        self.lr = self.get_current_lr() if self.get_current_lr() is not None else self.lr

        # This loop made for regularization loss calculation
        if self.regularizers:
            for layer in model.layers:
                for parameter in layer.trainable_params:
                    loss += mg.sum([reg.regularize(parameter)
                                   for reg in self.regularizers])

        loss.backward()
        if self.grad_clip_value is not None:
            all_grads = [p.grad for layer in model.layers for p in layer.trainable_params if p.grad is not None]
            if all_grads: # Ensure there are gradients to process
                # Calculate the L2 norm of all gradients combined
                norm = np.sqrt(sum(np.sum(g**2) for g in all_grads))
                if norm > self.grad_clip_value:
                    # Scale all gradients down if the norm is too high
                    scaling_factor = self.grad_clip_value / norm
                    for grad_tensor in all_grads:
                        grad_tensor *= scaling_factor
        for layer in model.layers:
            for parameter in layer.trainable_params:
                if parameter.grad is None:
                    continue
                self.optimize_param(parameter)


class SGD(Optimizer):
    """
    Stochastic Gradient Descend is a powerful optimiser and
        is more stable than Adam numerically
    """

    def __init__(self, lr, decay=None, momentum=None,
                 regularizers: List[Regularizer] = None, clip=None) -> None:
        """
        :param float lr: The learn rate of the optimiser
        :param float decay: The rate of learn rate decay can be ``None`` or
            ``False`` in order to avoid decay
        :param float momentum: The momentum but can be ``None`` or
            ``False`` in order to avoid decay
        :param list regularizers: The list which contains the regularizers
        """
        self.lr = self.init_lr = lr
        self.decay = decay
        self.momentum = momentum
        if self.momentum:
            self.momentums = {}
        super().__init__(lr, regularizers, clip)

    def get_current_lr(self):
        """
        :return: The learn rate with decay if existed
        :rtype: float

        This method returns the learn rate with respect to the current step
        """
        return self.init_lr * \
            (1. / (1 + self.decay * self.step)) if self.decay else self.lr

    def optimize_param(self, parameter: mg.Tensor) -> None:
        """
        :param mg.Tensor model: The model that needs to be trained

        This method performs training sequential models
        """
        self.lr = self.get_current_lr()
        if self.momentum:
            if id(parameter) not in self.momentums:
                self.momentums[id(parameter)] = np.zeros_like(parameter.data)
            param_update_value = ((self.momentum * self.momentums[id(parameter)]) -
                                  (self.lr * parameter.grad))
            self.momentums[id(parameter)] = param_update_value
            parameter.data += param_update_value
        else:
            parameter.data -= self.lr * parameter.grad


class Adam(Optimizer):
    """
    Adam a very good optimiser can converge quickly but less stable numerically

    :param float lr: The learn rate of the optimiser
    :param float decay: The rate of learn rate decay can be
        ``None`` in order to avoid decay
    """

    def __init__(self, learning_rate=0.001, decay=0., epsilon=1e-7,
                 beta_1=0.9, beta_2=0.999,
                 regularizers: List[Regularizer] = None, clip=None) -> None:
        self.lr = self.init_lr = learning_rate
        self.decay = decay

        self.epilson = epsilon
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.momentum = {}
        self.cache = {}
        super().__init__(learning_rate, regularizers, clip)

    def get_current_lr(self) -> float:
        """
        This method returns the learn rate with respect to the current step

        :return: The learn rate with decay if existed
        :rtype: float
        """
        return self.init_lr * \
            (1. / (1 + self.decay * self.step)) if self.decay else self.lr

    def optimize_param(self, parameter: mg.Tensor) -> None:
        if id(parameter) not in self.momentum:
            self.momentum[id(parameter)] = np.zeros_like(parameter.data)
            self.cache[id(parameter)] = np.zeros_like(parameter.data)

        self.momentum[id(parameter)] = (self.beta_1 *
                                        self.momentum[id(parameter)] +
                                        (1 - self.beta_1) * parameter.grad)

        param_momentum_corrected = (self.momentum[id(parameter)]
                                    / ((1 - self.beta_1)**(self.step)))

        self.cache[id(parameter)] = (self.beta_2 * self.cache[id(parameter)]
                                     + (1 - self.beta_2) * parameter.grad**2)

        param_cache_corrected = (self.cache[id(parameter)]
                                 / (1 - self.beta_2 ** (self.step)))

        parameter.data += -self.lr * param_momentum_corrected / \
            (np.sqrt(param_cache_corrected) + self.epilson)


class L1(Regularizer):
    """A L1 regularizer"""

    def __init__(self, lambda_: float):
        """
        :param float lambda_: The rate of penalty
        """
        self.lambda_ = lambda_
        super().__init__()

    def regularize(self, parameter: mg.Tensor) -> mg.Tensor:
        return self.lambda_ * mg.sum(mg.abs(parameter))


class L2(Regularizer):
    """A L2 regularizer"""

    def __init__(self, lambda_: float):
        """
        :param float lambda_: The rate of penalty
        """
        self.lambda_ = lambda_
        super().__init__()

    def regularize(self, parameter: mg.Tensor) -> mg.Tensor:
        return self.lambda_ * mg.sum(mg.power(parameter, 2))


class RMSProp(Optimizer):
    """Root Mean Square Propagation optimiser or for short named RMSProp"""

    def __init__(self, lr=0.001, decay=None, epsilon=1e-7, rho=0.9,
                 regularizers: List[Regularizer] = None):
        """
        :param float lr: The learning rate of optimizer
        :param float decay: The decay rate of the learning rate
            can be ``None`` to stop the decay
        :param float epsilon: The epsilon of the optimizer
        :param float rho: The rho of the optimizer
        :param List[:class:`helixnet.optimizers.Regularizer`] regularizers:
            The list which contains the regularizers that will regularize
            the parameters of the model
        """
        super().__init__(lr, regularizers)
        self.decay = decay
        self.epsilon = epsilon
        self.rho = rho
        self.cache = {}

    def get_current_lr(self):
        # Update the learning rate based on the current step/iteration
        if self.decay:
            return self.lr * (1.0 / (1.0 + self.decay * self.step))
        return self.lr

    def optimize_param(self, parameter: mg.Tensor) -> None:
        """This method contains the update logic for a single parameter."""
        # Initialize the cache for this parameter if it's the first time we've seen it.
        # Using id(parameter) is a robust way to get a unique key.
        if id(parameter) not in self.cache:
            self.cache[id(parameter)] = np.zeros_like(parameter.data)

        # --- The Core RMSProp Logic ---
        # 1. Update the cache with the exponentially weighted average of squared gradients
        self.cache[id(parameter)] = (self.rho * self.cache[id(parameter)] +
                                     (1 - self.rho) * parameter.grad**2)

        # 2. Update the parameter's data using the cache
        parameter.data += -self.lr * \
            parameter.grad / \
            (np.sqrt(self.cache[id(parameter)]) + self.epsilon)


class NesterovSGD(Optimizer):
    """
    A Nesterov Stochastic Gradient Descend optimizer.

    :param float lr: The learn rate of the optimiser
    :param float decay: The rate of learn rate decay can be ``None`` or
        ``False`` in order to avoid decay
    :param float momentum: The momentum value
    :param list regularizers: The list which contains the regularizers
    """

    def __init__(self, lr, decay=None, momentum=0.9,
                 regularizers: List[Regularizer] = None, clip=None) -> None:
        self.lr = self.init_lr = lr
        self.decay = decay
        self.momentum = momentum
        if self.momentum:
            self.momentums = {}
        super().__init__(lr, regularizers, clip)

    def get_current_lr(self):
        """
        :return: The learn rate with decay if existed
        :rtype: float

        This method returns the learn rate with respect to the current step
        """
        return self.init_lr * \
            (1. / (1 + self.decay * self.step)) if self.decay else self.lr

    def optimize_param(self, parameter: mg.Tensor) -> None:
        """
        :param mg.Tensor model: The model that needs to be trained

        This method performs training sequential models
        """
        if id(parameter) not in self.momentums:
            self.momentums[id(parameter)] = np.zeros_like(parameter.data)

        # Store the previous momentum update
        prev_momentum = self.momentums[id(parameter)]

        current_momentum = self.momentum * prev_momentum - self.get_current_lr() \
            * parameter.grad
        self.momentums[id(parameter)] = current_momentum

        parameter.data += -self.momentum * prev_momentum + (1 + self.momentum) * \
            current_momentum



import mygrad as mg
import numpy as np


def precision(y_pred, y_true):
    """Calculates precision for binary classification."""
    true_positives = mg.sum((y_pred == 1) & (y_true == 1))
    false_positives = mg.sum((y_pred == 1) & (y_true == 0))
    return true_positives / (true_positives + false_positives + 1e-7)  # Epsilon for stability


def recall(y_pred, y_true):
    """Calculates recall for binary classification."""
    true_positives = mg.sum((y_pred == 1) & (y_true == 1))
    false_negatives = mg.sum((y_pred == 0) & (y_true == 1))
    return true_positives / (true_positives + false_negatives + 1e-7)


def f1_score(y_pred, y_true):
    """Calculates the F1-score."""
    p = precision(y_pred, y_true)
    r = recall(y_pred, y_true)
    return 2 * (p * r) / (p + r + 1e-7)


def top_k_accuracy(y_pred_logits, y_true, k=5):
    """Calculates top-k accuracy.
    
    Args:
        y_pred_logits (np.ndarray): The raw logit outputs from the model.
        y_true (np.ndarray): The true integer labels.
        k (int): The number of top predictions to consider.
    """
    # Get the indices of the top k predictions for each sample
    top_k_preds = np.argsort(y_pred_logits, axis=1)[:, -k:]

    # Check if the true label is in the top k predictions for each sample
    matches = mg.any(top_k_preds == y_true.reshape(-1, 1), axis=1)

    return mg.mean(matches)


def rmse(y_pred, y_true):
    """
    It's just the square root of MSE. Its advantage is that the units are the same as the target
    """
    return np.sqrt(np.mean((y_pred - y_true)**2))


def r_squared(y_pred, y_true):
    """
    A standard statistical measure that represents
    the proportion of the variance in the dependent variable that is predictable from
    the independent variables. An R² of 1 indicates that the model explains
    all the variability of the response data around its mean.
    """
    ss_total = np.sum((y_true - np.mean(y_true))**2)
    ss_residual = np.sum((y_true - y_pred)**2)
    return 1 - (ss_residual / (ss_total + 1e-7))


def accuracy(y_pred, y_true):
    """Calculates the accuracy of the models"""
    # Assumes y_pred are logits or probabilities
    predictions = np.argmax(y_pred, axis=1)
    return np.mean(predictions == y_true)



import pathlib as pt
from typing import List, Dict
from PIL import Image
import numpy as np

def load_image(file_path:pt.Path):
    """
    This function loads the images in (channels, length, width) format
    in order to compatible with :class:`helixnet.layers.Conv2D`

    :param file_path: the path of the image to be loaded can a string or pathlib.Path object
    """
    with Image.open(file_path) as file:
        image = np.array(file).transpose(2, 0, 1)
    return image


import json
import mygrad as mg
import numpy as np
from . import models, layers, activations

LAYER_REGISTRY = {cls.__name__: cls for cls in layers.Layer.__subclasses__()}
ACTIVATION_REGISTRY = {}

for key, func in activations.__dict__.items():
    if callable(func):
        ACTIVATION_REGISTRY[func.__name__] = func


def save_model(model: models.Sequential, filepath: str):
    """
    Saves a model by cleanly separating architecture and weights.

    :param models.Sequential model: The model that will be saved
    :param str filepath: The filename of the model that should be saved in
    """
    model_data = {
        # The architecture is just a list of configs
        "architecture": [layer.get_config() for layer in model.layers],
        # The weights are just a list of numpy arrays
        "weights": [w.tolist() for layer in model.layers for w in layer.get_weights()]
    }
    with open(filepath, 'w') as f:
        json.dump(model_data, f, indent=4)


def load_model(filepath: str) -> models.Sequential:
    """
    Loads a model from its architecture and weights.

    :param str filepath: The filename of the model that will be loaded from.
    """
    with open(filepath, 'r') as f:
        model_data = json.load(f)

    # 1. Rebuild the empty model from the architecture blueprint
    loaded_layers = []
    for layer_config in model_data['architecture']:
        class_name = layer_config.pop('class_name') # Assumes get_config adds this
        LayerClass = LAYER_REGISTRY[class_name]

        # Restore the activation function from the registry
        if 'activation' in layer_config:
            layer_config['activation'] = ACTIVATION_REGISTRY.get(layer_config['activation'])

        # Create the layer!
        layer = LayerClass(**layer_config)
        loaded_layers.append(layer)

    model = models.Sequential(loaded_layers)

    # 2. Load the saved weights into the empty model
    all_weights = [np.array(w) for w in model_data['weights']]
    model.set_weights(all_weights)

    return model




"""This module contains model creation capablities and tools"""
from typing import List, Tuple, Dict
from functools import partial


import mygrad as mg
import numpy as np

from typing import List, Tuple, Dict, Callable, Optional # Make sure these are present
from rich.live import Live
from rich.progress import (Progress, SpinnerColumn, BarColumn, TextColumn, TimeRemainingColumn,
                           TaskProgressColumn, track)
from rich.table import Table
from rich.console import Group
from rich.panel import Panel
from rich import print


from helixnet import layers


class Sequential:
    """A Simple model that propagate through the layers in a linear way

    :param list layer: the list which contains the layers"""

    def __init__(self, layers_: list[layers.Layer]) -> None:
        self.layers = layers_

    def forward(self, x: mg.Tensor) -> mg.Tensor:
        """Perform a prediction across multiple layers

        Args:
            x (mg.tensor): the input

        Returns:
            mg.tensor: the predictions
        """
        for layer in self.layers:
            x = layer.forward(x)
        return x

    def get_names(self) -> List[str]:
        """Returns a list of layers names

        Returns:
            list: A list of strings
        """
        return [layer.name for layer in self.layers]

    def null_grads(self) -> None:
        """Reset the gradients of every layer"""
        for layer in self.layers:
            layer.null_grad()

    def output_shape(self) -> Tuple[int]:
        """A simple function that shows the model's last layer's output shape"""

        try:
            shape = tuple()
            for layer in self.layers:
                shape = layer.output_shape(shape)
            return shape
        except Exception as e:
            raise Exception(f"An Error occurred at {layer.name=} with {shape=}\n"
                            f"Err: {e.args}")

    def add(self, layer: layers.Layer) -> None:
        """This function can append layers to the model

        :param layer.Layer layer: The layer that will be appended to the end of the model"""
        self.layers.append(layer)

    def summary(self) -> None:
        """
        This method prints the model summary which contains
        the name of every layer and it's shape
        """
        table = Table(title="The Model Summary")
        table.add_column(" ", ratio=1)
        table.add_column("Layer", ratio=5)
        table.add_column("Output Shape", ratio=5)
        table.add_column("No. Params", ratio=3)
        shape = []
        total_params = 0
        for i, layer in enumerate(self.layers):
            shape = layer.output_shape(shape)
            params = layer.total_params()
            total_params += params
            table.add_row(str(i), f"{layer.name} ({layer.type})", str(shape), str(params))
        table.add_section()
        print(table)
        print("Total Parameters", str(total_params))

    def predict(self, x: mg.Tensor) -> mg.Tensor:
        """This method let the model predict without building computational graph

        :param mg.Tensor x: The models input
        :return mg.Tensor: The models predictions"""
        for layer in self.layers:
            x = layer.predict(x)
        return x

    def get_weights(self) -> List[np.ndarray]:
        """Returns a flat list of all trainable weights in the model."""
        return [weight for layer in self.layers for weight in layer.get_weights()]

    def set_weights(self, weights: List[np.ndarray]):
        """
        Sets the model's weights from a flat list.

        :param List[np.ndarray] weights: The weights what will be produced by
            :func:`helixnet.io.load_model`
        """
        weight_iterator = iter(weights)
        for layer in self.layers:
            # Slicing the iterator to get the correct number of weights for this layer
            num_params = len(layer.trainable_params)
            layer_weights = [next(weight_iterator) for _ in range(num_params)]
            layer.set_weights(layer_weights)


    def fit(self, X, Y, loss_func: Callable, optimizer, epochs: int = 1,
            batch_size: Optional[int] = None,
            preprocessing: Optional[Callable] = None,
            metrics: Optional[Dict[str, Callable]] = None):
        """
        A high-level training loop with a rich, interactive display.
        """
        # --- Full-Batch Training (Simpler UI) ---
        if batch_size is None:
            print("[bold cyan]Starting full-batch training...[/bold cyan]")
            for epoch in track(range(epochs), description="Training epochs"):
                x_processed = X if not preprocessing else preprocessing(X)
                prediction = self.forward(x_processed)
                loss = loss_func(prediction, Y)
                optimizer.optimize(self, loss)
                if epoch % 100 == 0 or epoch == epochs - 1:
                    log_line = f"Epoch: {epoch}, Loss: {loss.item():.4f}"
                    if metrics:
                        for name, func in metrics.items():
                            metric_val = func(prediction.data, Y)
                            log_line += f", {name.title()}: {metric_val:.4f}"
                    print(log_line)
                optimizer.epoch_done()
            print("[bold green]Full-batch training complete.[/bold green]")
            return

        overall_progress = Progress(
            TextColumn("[bold red]Overall Progress", justify="right"),
            BarColumn(bar_width=None),
            TaskProgressColumn(),
            "•",
            TimeRemainingColumn(), # Stable ETA for the entire training run
        )

        batch_progress = Progress(
            # We will update its content dynamically every batch.
            TextColumn("{task.description}"),
            BarColumn(bar_width=None),
            TaskProgressColumn(),
            "•",
            TimeRemainingColumn(), # ETA for the current epoch
        )

            # Define the results table
        results_table = Table(title="Training Log", show_header=True, header_style="bold magenta")
        results_table.add_column("Epoch", justify="right", style="cyan")
        results_table.add_column("Avg Train Loss", justify="center", style="green")
        if metrics:
            for name in metrics.keys():
                results_table.add_column(f"Avg Train {name.replace('_', ' ').title()}", justify="center")

        # Group UI elements inside a Panel
        render_group = Panel(
           Group(results_table, batch_progress, overall_progress),
           title="[bold yellow]HelixNet Training[/bold yellow]",
           border_style="blue"
            )

        with Live(render_group, refresh_per_second=10, vertical_overflow="visible") as live:
            num_batches_per_epoch = (len(X) + batch_size - 1) // batch_size
            total_batches = num_batches_per_epoch * epochs
            overall_task = overall_progress.add_task("Batches", total=total_batches)

            for epoch in range(epochs):
                epoch_loss_total = 0.0
                epoch_metrics_totals = {name: 0.0 for name in metrics.keys()} if metrics else {}

                # Add the batch task with a simple initial description
                batch_task = batch_progress.add_task(f"Epoch {epoch+1}", total=num_batches_per_epoch)

                batch_indices = np.arange(len(X))
                np.random.shuffle(batch_indices)

                for i in range(num_batches_per_epoch):
                    start, end = i * batch_size, (i + 1) * batch_size
                    # ... (batch creation logic) ...
                    x_batch, y_batch = X[start:end], Y[start:end]

                    # ... (forward pass, optimizer.optimize, etc.) ...
                    prediction = self.forward(x_batch)
                    loss = loss_func(prediction, y_batch)
                    optimizer.optimize(self, loss)

                    current_loss = loss.item()
                    epoch_loss_total += current_loss
                    if metrics:
                        for name, func in metrics.items():
                            epoch_metrics_totals[name] += func(prediction.data, y_batch)

                    # This is the most reliable method.
                    live_description = (f"[bold cyan]Epoch {epoch+1}[/bold cyan] "
                                        f"[green]Loss: {current_loss:.4f}[/green]")

                    batch_progress.update(batch_task, advance=1, description=live_description)
                    overall_progress.update(overall_task, advance=1)

                # --- After an epoch is complete ---
                avg_loss = epoch_loss_total / num_batches_per_epoch
                row_data = [f"{epoch + 1}", f"{avg_loss:.4f}"]
                if metrics:
                    for name in metrics.keys():
                        avg_metric = epoch_metrics_totals[name] / num_batches_per_epoch
                        row_data.append(f"{avg_metric:.4f}")

                results_table.add_row(*row_data)

                # Remove the completed batch task
                batch_progress.remove_task(batch_task)

                optimizer.epoch_done()



import numpy as np
import mygrad as mg


def CatCrossEntropy(y_pred, y_true):
    """
    Performs categorical cross-entropy. Useful for classification
    """
    return -(mg.sum(y_true * mg.log(y_pred + 1e-8))) / len(y_pred)


def MeanSquaredError(y_pred, y_true):
    """
    Mean Squared error loss. Useful for regression
    """
    return mg.mean((y_pred - y_true) * (y_pred - y_true))


def MeanAbsError(y_pred, y_true):
    """
    Mean Absolute error loss. Useful for regression
    """
    return mg.mean(mg.abs(y_pred - y_true))


def Kullback_Leibler(y_pred, y_true):
    return y_pred * mg.log(y_pred / y_true) + (1 - y_pred) * \
        mg.log((1 - y_pred) / (1 - y_true))


def softmax_crossentropy(y_pred, y_true):
    """
    Performs categorical cross-entropy but for models that use softmax as
    their last activation function. However it must accept the raw logits
    (without applying softmax for the last layers).
    Useful for classification
    """
    return mg.nnet.losses.softmax_crossentropy(y_pred.astype(np.int32), y_true.astype(np.int32))


def HuberLoss(y_pred, y_true, delta=1.0):
    """
    A great general-purpose regression loss that is less sensitive to outliers than MSE.
    It's a fantastic default choice for many regression problems.
    """
    error = y_true - y_pred
    is_small_error = mg.abs(error) <= delta

    squared_loss = 0.5 * (error**2)
    linear_loss = delta * (mg.abs(error) - 0.5 * delta)

    return mg.where(is_small_error, squared_loss, linear_loss)


def LogCoshLoss(y_pred, y_true):
    """A Very smooth loss function for regression that is less sensitive to outliers than MSE."""
    error = y_true - y_pred
    return mg.log(mg.cosh(error))


def FocalLoss(y_pred_sigmoid, y_true, alpha=0.25, gamma=2.0):
    """
    **Crucial for imbalanced datasets** in binary or
    multi-class classification (e.g., object detection where 
    "background" is the most common class).
    """
    p_t = y_pred_sigmoid * y_true + (1 - y_pred_sigmoid) * (1 - y_true)
    alpha_t = alpha * y_true + (1 - alpha) * (1 - y_true)

    bce = -mg.log(p_t)
    focal_modulator = (1 - p_t)**gamma

    return alpha_t * focal_modulator * bce


def HingeLoss(y_pred, y_true):
    """
    Primarily for "max-margin" classification, famously
    used in Support Vector Machines (SVMs). It's great
    when you want to train a model that doesn't just classify
    correctly but does so with a high degree of confidence.

    **Predictions should be in the [-1, 1] range.**
    """
    # y_true should be -1 or 1
    return mg.maximum(0, 1 - y_true * y_pred)


def CosineSimilarityLoss(y_pred, y_true):
    """
    When the direction of the embedding vectors is
    more important than their magnitude.
    It's very common in NLP for comparing sentence embeddings
    """
    # Normalize the vectors to unit length
    y_pred_norm = y_pred / mg.sqrt(mg.sum(y_pred**2, axis=-1, keepdims=True))
    y_true_norm = y_true / mg.sqrt(mg.sum(y_true**2, axis=-1, keepdims=True))

    # Calculate cosine similarity
    cosine_similarity = mg.sum(y_pred_norm * y_true_norm, axis=-1)

    # The loss is 1 minus the similarity
    return 1 - cosine_similarity


def BinaryCrossEntropy(y_pred, y_true):
    """Binary cross-entropy useful for classification"""
    return -(y_true * mg.log(y_pred) + (1 - y_true) * mg.log(1 - y_pred))
